from .component import (
    Component,
    JinjaFileTemplateBase,
    JinjaRelativeFileTemplateBase,
    JinjaStringTemplate,
    StringTemplate,
)
from .decorators import (
    dataclass_component,
    dataclass_swappable_component,
    template_field,
)

__all__ = [
    "Component",
    "JinjaStringTemplate",
    "StringTemplate",
    "JinjaFileTemplateBase",
    "JinjaRelativeFileTemplateBase",
    "dataclass_component",
    "dataclass_swappable_component",
    "template_field",
]



import functools
import typing as t
from dataclasses import (
    _FIELDS as _DATACLASS_FIELDS,  # type: ignore
)
from dataclasses import (
    _MISSING_TYPE,  # pyright: ignore
    Field,
    dataclass,
    field,
    fields,
)

from .component import Component

if t.TYPE_CHECKING:
    from _typeshed import DataclassInstance


_DATACLASS_COMPONENT_SWAPPABLE = "__is_dataclass_component__"


def is_missing(val: t.Any) -> bool:
    return val is _MISSING_TYPE or isinstance(val, _MISSING_TYPE)


def is_required(field: Field[t.Any]) -> bool:
    """Checks whether a dataclass field is required or not."""
    if not is_missing(field.default):
        return False
    if not is_missing(field.default_factory):  # type: ignore
        return False
    if field.init is True:
        return True
    return False


def required_fields(cls: "DataclassInstance") -> list[Field[t.Any]]:
    return [f for f in fields(cls) if is_required(f)]


def is_dataclass_swappable_component(obj: object) -> bool:
    return hasattr(obj, _DATACLASS_COMPONENT_SWAPPABLE)


def check_swappable_fields(cls: "DataclassInstance"):
    for field_ in fields(cls):
        origin = t.get_origin(field_.type)
        args = t.get_args(field_.type)
        # If the field was defined as: type[Component]
        if origin is type:
            type_arg = args[0]

            if issubclass(type_arg, Component) and not is_dataclass_swappable_component(
                type_arg
            ):
                raise TypeError(
                    f"""In {cls}, field `{field_.name}: {field_.type}` is not valid because the class {type_arg} is not swappable. Please decorate the class (or any parent) with @dataclass_swappable_component."""
                )


@functools.wraps(dataclass)
def _dataclass_component(  # type: ignore[reportUnusedFunction]
    cls: t.Type[object] | None = None, /, swappable: bool = False, **kwargs: t.Any
):
    """Wrapper around dataclass that turns a class into a dataclass component.
    If swappable=True, enforces that child classes adhere to parent component's interface (i.e. has the same initialization interface).
    """
    # Handles case of calling with parenthesis: @dataclass(...)
    if cls is None:
        # Call this class again with a closure where the `cls` is first converted to a dataclass called with kwargs
        return lambda cls: _dataclass_component(dataclass(**kwargs)(cls))  # type: ignore[reportUnknownVariableType]

    # Handles case of calling without parenthesis: @dataclass
    if _DATACLASS_FIELDS not in cls.__dict__:
        cls = dataclass(cls)

    # Check if any fields are t.Type[Component] and validate it is a swappable component
    check_swappable_fields(cls)  # type: ignore

    # Need to dynamically modify the existing class to specify it is a dataclass_component.
    if swappable:
        setattr(cls, _DATACLASS_COMPONENT_SWAPPABLE, True)

    # Get the field names
    cls_fields = {f.name for f in required_fields(cls)}  # type: ignore

    # If a parent class is a dataclass, check that no extra required fields are present in the child class.
    # It's ok to only check direct with `__bases__` and not `__mro__` since fields(base) will contain all ancestors fields.
    for base in cls.__bases__:
        if is_dataclass_swappable_component(base):  # true if any ancestor is swappable
            # A class will have all fields of its ancestors plus any new fields it defined
            allowed_fields = {f.name for f in fields(base)}
            extra_fields = cls_fields - allowed_fields
            if extra_fields:
                raise TypeError(
                    f"Extra required attributes not allowed in class `{cls.__name__}` when subclassing `{base.__name__}`: {', '.join(extra_fields)}."
                    "\nChild components must be consistent with the parent's attributes in order to have a composable interface."
                )
    return cls


# There are special stubs for dataclass that are difficult to recreate so for typechecking just use dataclass

dataclass_component = dataclass if t.TYPE_CHECKING else _dataclass_component
dataclass_swappable_component = (
    dataclass
    if t.TYPE_CHECKING
    else functools.partial(_dataclass_component, swappable=True)
)


# Component Dataclass Helpers
template_field = functools.partial(field, init=False, repr=False, compare=False)



import inspect
import os
import typing as t
from copy import copy
from dataclasses import (
    fields,
    is_dataclass,
)
from inspect import isclass

from jinja2 import BaseLoader, Environment, StrictUndefined
from jinja2.loaders import FileSystemLoader


# An isinstance check that only checks for object instances (returns False for subclsases)
def is_component_instance(value: t.Any):
    return isinstance(value, Component) and not (
        isclass(value) and issubclass(value, Component)
    )


def recursively_render(obj: t.Any) -> t.Any:
    if is_component_instance(obj):
        return obj.render()
    if isinstance(obj, list):
        return [recursively_render(item) for item in obj]  # type: ignore[reportUnknownVariableType]
    if isinstance(obj, tuple):
        return tuple(recursively_render(item) for item in obj)  # type: ignore[reportUnknownVariableType]
    if isinstance(obj, dict):
        return {k: recursively_render(v) for k, v in obj.items()}  # type: ignore[reportUnknownVariableType]
    return obj


def render_component_vars(component: "Component") -> dict[t.Any, t.Any]:
    """Makes a shallow copy of the component and recursively calls `.render()` on fields that are Components.
    The default `asdict` implementation has no way to provide custom serialization because `Component` is a dataclass, and by default
    it will it simply render its fields."""
    # Shallow copy into a new dataclass object, to avoid overwriting fields
    component = copy(component)
    #  Calls class method
    component._pre_render(component)  # pyright: ignore[reportPrivateUsage]

    if not is_dataclass(component):
        raise ValueError(f"Expected dataclass received {component}")

    # Recursively render fields and add them to dict
    rendered_vars: dict[str, t.Any] = {}
    for field_ in fields(component):
        value = getattr(component, field_.name)
        value = recursively_render(value)
        if isinstance(value, Component):
            value = value.render()  # Recursively render.
        rendered_vars[field_.name] = value

    exported_vars = component._post_render(rendered_vars)  # pyright: ignore[reportPrivateUsage]
    return exported_vars


@t.runtime_checkable
class Component(t.Protocol):
    """This class is a protocol to deal with the fact that dataclasses do not respect `is_instance` checks for inheritance.
    We must use runtime_checkable to check for subclasses of `Component`.

    Use the @dataclass_component on your components.

    Exported fields follow the rules of dataclasses:
     - underscore variables are private
     - t.ClassVar variables are private
     - dynamically set attributes are not exported

     There are 3 hooks:
      - __post_init__: Modify the original instance.
      - _pre_render: Modify a shallow copy of the dataclass before it's rendered.
      - _post_render: Modify the rendered dict that will be sent as variables to the template.

    Prefer `_pre_render` over `__post_init__` since it doesn't modify the original state of the object.
    _pre_render will not be stale (like post_init) if attributes are mutated.
    """

    def __post_init__(self):
        """Modify to run a custom init for dataclasses, see https://docs.python.org/3/library/dataclasses.html#dataclasses.__post_init__"""
        pass

    @classmethod
    def _pre_render(cls, self: t.Self):
        """Operates on a shallow copy of `self` before rendering.
        Modify existing attributes on the component to change them.
        This avoids modifying the original object."""
        pass

    @classmethod
    def _post_render(cls, template_vars: dict[str, t.Any]):
        """Postprocessing before the dictionary is sent to the template"""
        return template_vars

    def render(self) -> str:
        raise NotImplementedError()


DEFAULT_JINJA_ENV = Environment(loader=BaseLoader(), undefined=StrictUndefined)


class JinjaStringTemplate(Component, t.Protocol):
    """This class can be subclassed to override the jinja environment.
    A default environment for rendering strings is used."""

    _template: t.ClassVar[str]
    _jinja_environment: t.ClassVar[Environment] = DEFAULT_JINJA_ENV

    def render(self) -> str:
        template = self._jinja_environment.from_string(self._template)
        return template.render(**render_component_vars(self))  # pyright: ignore[reportArgumentType]


class StringTemplate(Component, t.Protocol):
    _template: t.ClassVar[str]

    def render(self) -> str:
        return self._template.format(**render_component_vars(self))  # pyright: ignore[reportArgumentType]


class JinjaFileTemplateBase(Component, t.Protocol):
    """This class is intended to be subclass, first with the environment set as a class var. E.g.
    ```
    my_environment = SandboxedEnvironment(...)

    class JinjaFileTemplate(JinjaFileTemplateBase):
        _jinja_environment: my_environment

    class MyTemplate(JinjaFileTemplate):
        _template = "path_to_template"
        var_a = "a"
    ```

    """

    _template_path: t.ClassVar[str]
    _jinja_environment: t.ClassVar[Environment]

    def _get_template_path(self) -> str:
        return self._template_path

    def render(self) -> str:
        template = self._jinja_environment.get_template(self._get_template_path())
        return template.render(**render_component_vars(self))  # pyright: ignore[reportArgumentType]


class JinjaRelativeFileTemplateBase(JinjaFileTemplateBase, t.Protocol):
    """
    Allows for defining templates as relative file paths to the current file.
    Absolute paths still must exist in the FileSystemLoader path(s) provided in the environment config.

    Subclasses must define:
      - _jinja_environment: an Environment with a FileSystemLoader
      - _template_path: a relative path to the template file (relative to the file where the subclass is defined)

    Example:

        my_environment = SandboxedEnvironment(
            loader=FileSystemLoader(searchpath=PROMPT_DIR),
            undefined=jinja2.StrictUndefined,
        )

        class JinjaFileTemplate(JinjaRelativeFileTemplateBase):
            _jinja_environment = my_environment

        # In some other file
        class MyTemplate(JinjaFileTemplate):
            _template_path = "relative_path_to_template.jinja2"
            var_a = "a"
    """

    def __init_subclass__(cls, **kwargs: t.Any):
        super().__init_subclass__(**kwargs)
        # Check that _jinja_environment is defined and its loader is a FileSystemLoader.
        env = getattr(cls, "_jinja_environment", None)
        if env is not None:
            loader = getattr(env, "loader", None)
            if not isinstance(loader, FileSystemLoader):
                raise TypeError(
                    "The Jinja environment's loader must be an instance of FileSystemLoader"
                )

    def _get_template_path(self) -> str:
        """
        Computes the template path relative to one of the loader's search paths.

        1. Determine the absolute path of the template based on where the subclass is defined.
        2. Iterate over all search directories in the loader.
        3. For the first search directory that is a prefix of the template's absolute path,
           compute and return the relative path.
        4. If none match, raise an error.
        """
        relative_template_path = super()._get_template_path()
        # Get the file where the subclass is defined.
        subclass_file = inspect.getfile(self.__class__)
        base_dir = os.path.dirname(os.path.abspath(subclass_file))
        abs_template_path = os.path.join(base_dir, relative_template_path)
        loader = t.cast(FileSystemLoader, self._jinja_environment.loader)
        # Iterate over all search directories
        for search_dir in loader.searchpath:
            norm_search_dir = os.path.abspath(search_dir)
            if abs_template_path.startswith(norm_search_dir):
                return os.path.relpath(abs_template_path, start=norm_search_dir)

        raise ValueError(
            f"Template '{abs_template_path}' is not within any of the search paths: "
            f"{loader.searchpath}"
        )
